# -*- coding: utf-8 -*-
"""
Created on Wed Dec 02 20:29:59 2020

@author: Theano Iliopoulou
Code citation: T. Iliopoulou, and D. Koutsoyiannis, PythOm: A python toolbox implementing recent advances in rainfall intensity (ombrian) curves, EGU General Assembly 2021, online, EGU21-389, doi:10.5194/egusphere-egu21-389, European Geosciences Union, 2021.
© Creative Commons Attribution 4.0 International (CC BY 4.0) 

"""

from __future__ import division
from math import *
import pandas as pd
import numpy as np
from scipy.special import gamma, gammaln

## climacogram models
# Filtered HK Cauchy type 
def FHK_C(k,l1,a,M,H):
    g_k=l1*((1+(k/a)**(2*M))**((H-1)/M))
    return g_k

## Filtered HK Cauchy-Dagum-type (FHK-CD) for M=1-H (rough process)
def FHK_CD(k,l1,l2,a,H):
    g_k=l1*(1+k/a)**(2*H-2)+l2*(1-(1+a/k)**(2*H-2))
    return g_k


## K-moment for order p
def kmom(p,x): ## Non-central Unbiased for q=1, x must be sorted in increasing order
    x=x[~np.isnan(x)]
    n=len(x)
    b=np.empty(n)
    b[:]=np.NaN
    i=np.arange(1,n+1,1)
    i_0=i[i<p]
    i_1=i[i>=p]
    b[i_0-1]=0
    b[i_1-1]=(p/n)*np.exp(gammaln(n-p+1)-gammaln(n)+gammaln(i_1)-gammaln(i_1-p+1)) 
    kmom=np.sum(b*x)
    return kmom  


## Return periods for K-moments
def K_return_periods(params,s_len,d_len,k_trans,k_h,scales_h,scales_d,Cg_type,pw_k,p): 
    m=params[0]
    th=params[1]
    xi=params[6]
    l1=params[2]
    scales=scales_h+scales_d
       
    if Cg_type=='CD':
        l2=params[3]
        a=params[4]
        H=params[5]
    ## theoretical climacogram in hours   
        th_var_k=pd.Series((FHK_CD(k_h,l1,l2,a,H)),index=scales)
        th_var_L_h=FHK_CD(s_len*k_h[0],l1,l2,a,H) # bias for hourly series
        th_var_L_d=FHK_CD(d_len*24,l1,l2,a,H)     # bias for daily series (in hours)
    elif Cg_type=='C':
         a=params[3]
         M=params[4]
         H=params[5]
    ## theoretical climacogram in hours   
         th_var_k=pd.Series((FHK_C(k_h,l1,a,M,H)),index=scales)
         th_var_L_h=FHK_C(s_len*k_h[0],l1,a,M,H) # bias for hourly series
         th_var_L_d=FHK_C(d_len*24,l1,a,M,H)     # bias for daily series (in hours)    
    ## theoretical probability wet
    k_trans_idx0=np.where(k_h == k_trans)       #index of k_trans in scales and k_h (second index occurrence)
    if len(k_trans_idx0)>1:
       k_trans_idx=k_trans_idx0[0][1]
    elif len(k_trans_idx0)==1:
        k_trans_idx=k_trans_idx0[0][0]
    th_pw_k_P=((1-xi)/(0.5-xi))*((m**2)/(th_var_k+m**2)) # for Pareto scales
    th_pw_k_trans=th_pw_k_P[k_trans_idx]                 # for transition scale
    th_pw_k_PBF=pd.Series(1-(1-th_pw_k_trans)**(k_h/k_trans)**th, index=scales)        #for PBF scales
    th_pw_k=pd.concat([th_pw_k_P[:k_trans_idx+1],th_pw_k_PBF[k_trans_idx+1:]],axis=0)  # final model pw
  
    # function inverse of z_k (1/z_k)
    inv_z_k=((1-2*xi)*(th_pw_k*(th_var_k/m**2+1)-1))**0.5
 
    # coefficients L1, Linf for Pareto and PBF
    xi_2=xi*inv_z_k
    L1_P=(1-xi_2)**(-1/xi_2)    # for Pareto scales
#    L1=((beta(inv_z_k/xi_2-inv_z_k,inv_z_k)*inv_z_k)**(1/inv_z_k)+1)**(1/xi) #for PBF scales and model
    Linf=(gamma(1-xi_2))**(1/xi_2)
    
    # bias correction factor
    th_k=pd.DataFrame(np.empty([1,len(scales)]),columns=scales)
    th_k[:] = np.NaN
    
    for i in scales_h:
        th_k[i]=-th_var_L_h/(2*th_var_k[i])     
    for i in scales_d:
        th_k[i]=-th_var_L_d/(2*th_var_k[i])     
   
    # p_2 accounting for bias
    p_2= pd.DataFrame(np.empty([len(p),len(scales)])) # 
    p_2.columns=scales
    p_2[:] = np.NaN

    for i in scales:
        p_2[i][:]=2*th_k[i].values+(1-2*th_k[i].values)*p[i].values**((1+th_k[i].values)**2)
           
    k_y=np.array(k_h/(365.25*24))
    K_rp = pd.DataFrame(np.empty([len(p),len(scales)])) # set matrix with  theoretical return periods 
    K_rp[:] = np.NaN
    K_rp.columns=scales
    
    for i in np.arange(0,len(scales)):
        scale=scales[i]
        K_rp[scale][:]=(Linf[scale]*(p_2[scale].values-1)+L1_P[scale])*k_y[i]/pw_k[scale] 
    return K_rp

## 7-parameter ombrian model   
def ombrian_model(params,k_trans,k_h,scales,rp,Cg_type):
    m=params[0]                    # Mean intensity parameter
    th=params[1]                   # Exponent of the expression of probability dry parameter 
    xi=params[6]                   # Tail index parameter
    l1=params[2]                   # Intensity scale parameter (first)
    H=params[5]                    # Hurst parameter
    
    if Cg_type=='CD':
        l2=params[3]               # Intensity scale parameter (second)
        a=params[4]                # Time scale parameter
       
    ## theoretical climacogram in hours   
        th_var_k=pd.Series((FHK_CD(k_h,l1,l2,a,H)),index=scales)
       
    elif Cg_type=='C':
         a=params[3]               # Time scale parameter
         M=params[4]               # Fractal parameter
        
    ## theoretical climacogram in hours   
         th_var_k=pd.Series((FHK_C(k_h,l1,a,M,H)),index=scales)
        
    ## theoretical probability wet
    k_trans_idx0=np.where(k_h == k_trans)  #index of k_trans in scales and k_h (second index occurrence)
    if len(k_trans_idx0)>1:
       k_trans_idx=k_trans_idx0[0][1]
    elif len(k_trans_idx0)==1:
        k_trans_idx=k_trans_idx0[0][0]
    th_pw_k_P=((1-xi)/(0.5-xi))*((m**2)/(th_var_k+m**2))   # for Pareto scales
    th_pw_k_trans=th_pw_k_P[k_trans_idx]                   # for transition scale
    th_pw_k_PBF=pd.Series(1-(1-th_pw_k_trans)**(k_h/k_trans)**th, index=scales)         # for PBF scales
    th_pw_k=pd.concat([th_pw_k_P[:k_trans_idx+1],th_pw_k_PBF[k_trans_idx+1:]],axis=0)   # final model pw
    
    # function inverse of z_k (1/z_k)
    inv_z_k=((1-2*xi)*(th_pw_k*(th_var_k/m**2+1)-1))**0.5
    
    # function l_k
    l_k_P=m*(1-xi)/th_pw_k_P      # for Pareto scales
    l_k_PBF=m/th_pw_k_PBF/(1+inv_z_k**2/(1-xi)-inv_z_k**(2**0.5))           # for PBF scales
    l_k=pd.concat([l_k_P[:k_trans_idx+1],l_k_PBF[k_trans_idx+1:]],axis=0)   # final model pw
   
    ## theoretical quantiles
    k_y=np.array(k_h/(365.25*24))

    th_x = pd.DataFrame(np.empty([len(rp),len(scales)])) 
    th_x[:] = np.NaN
    th_x.columns=scales
    for i in np.arange(0,len(scales)):
        scale=scales[i]
        th_x[scale][:]=l_k[scale]*(((th_pw_k[scale]*rp[scale].values/k_y[i])**xi-1)/xi)**inv_z_k[scale]  # Ombrian model
 
    return th_x    

## Model error based on quantiles
def ombrian_quant_error(params,s_len,d_len,k_trans,k_h,scales_h,scales_d,Cg_type,pw_k,Ekmoms,p):
    scales=scales_h+scales_d
    l1=params[2]
    if Cg_type=='CD':
       l2=params[3]
       a=params[4]
       H=params[5]
    ## theoretical climacogram in hours   
       th_var_k=pd.Series((FHK_CD(k_h,l1,l2,a,H)),index=scales)
       
    elif Cg_type=='C':
         a=params[3]
         M=params[4]
         H=params[5]
    ## theoretical climacogram in hours   
         th_var_k=pd.Series((FHK_C(k_h,l1,a,M,H)),index=scales)
      
    K_rt=K_return_periods(params,s_len,d_len,k_trans,k_h,scales_h,scales_d,Cg_type,pw_k,p)
    th_x=ombrian_model(params,k_trans,k_h,scales,K_rt,Cg_type)

    
    ## Error in quantiles
    Etot=np.NaN
    if (th_x.isnull().sum().sum()>1 or params[1]>1 or H>0.95):
        Etot=10**8           # penalty for minimization when quantiles = nan and constraints violated (For Nelder-Mead method)
    elif th_x.isnull().sum().sum()==0:
        E= pd.DataFrame(np.empty([1,len(scales)])) # 
        E.columns=scales
        E[:] = np.NaN
        for i in scales:
            E[i]=np.nansum((Ekmoms[i].values-th_x[i].values)**2)
         
        # Standardized error by theoretical climacogram
        E_stand=E/th_var_k
        Etot0=E_stand.sum(axis=1) 
        Etot=Etot0.values
    return Etot  

## climacogram CD-type error   
def CgCD_Error(params,s_len,d_len,Ecg,k_h,scales_h,scales_d):
    l1=params[0]
    l2=params[1]
    a=params[2]
    H=params[3]
    scales=scales_h+scales_d
### theoretical and bias for CD  
    CD_var_k=pd.Series((FHK_CD(k_h,l1,l2,a,H)),index=scales)
    CD_var_L_h=FHK_CD(s_len*k_h[0],l1,l2,a,H)  # bias for hourly series
    CD_var_L_d=FHK_CD(d_len*24,l1,l2,a,H)      # bias for daily series (in hours)
   
## subtract bias
    CD_var_k2=CD_var_k
    for i in scales_h:
          CD_var_k2[i]=CD_var_k[i]-CD_var_L_h
    for i in scales_d:
         CD_var_k2[i]=CD_var_k[i]-CD_var_L_d
## Error 
    E_CD=(np.log(CD_var_k2)-np.log(Ecg ))**2
    
    EtotCD=E_CD.sum() 
    return EtotCD    

## climacogram C-type error   
def CgC_Error(params,s_len,d_len,Ecg,k_h,scales_h,scales_d):
    l1=params[0]
    a=params[1]
    M=params[2]
    H=params[3]    
    scales=scales_h+scales_d
 
### theoretical and bias for C    
    C_var_k=pd.Series((FHK_C(k_h,l1,a,M,H)),index=scales)
    C_var_L_h=FHK_C(s_len*k_h[0],l1,a,M,H)   # bias for hourly series
    C_var_L_d=FHK_C(d_len*24,l1,a,M,H)       # bias for daily series (in hours)
 
## subtract bias
    C_var_k2=C_var_k
    for i in scales_h:
          C_var_k2[i]=C_var_k[i]-C_var_L_h
    for i in scales_d:
          C_var_k2[i]=C_var_k[i]-C_var_L_d
## Error 
 
    E_C=(np.log(C_var_k2)-np.log(Ecg ))**2
    
    EtotC=E_C.sum() 
    return EtotC    

## Probability wet error
def pw_error(params,s_len,d_len,k_h,scales_h,scales_d,Cg_type,pw_k):
    m=params[0]
    xi=params[6]
    l1=params[2]
    scales=scales_h+scales_d
    if Cg_type=='CD':
        l2=params[3]
        a=params[4]
        H=params[5]
    ## theoretical climacogram in hours   
        th_var_k=pd.Series((FHK_CD(k_h,l1,l2,a,H)),index=scales)
       
    elif Cg_type=='C':
         a=params[3]
         M=params[4]
         H=params[5]
    ## theoretical climacogram in hours   
         th_var_k=pd.Series((FHK_C(k_h,l1,a,M,H)),index=scales)
        
    ## theoretical probability wet
    th_pw_k_P=((1-xi)/(0.5-xi))*((m**2)/(th_var_k+m**2)) 
    ## divergence:
    E_p=(th_pw_k_P-pw_k)**2
    E_totp=E_p.sum()
    return E_totp


